from __future__ import print_function, division
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.widgets import Slider
from agaro import output_utils
from bannock.utils import utils
from cannock.utils import utils as cutils


def plot_2d(dirname):
    fig = plt.figure()
    ax_vis = fig.add_subplot(111)

    fnames = output_utils.get_filenames(dirname)
    m_0 = output_utils.filename_to_model(fnames[0])

    L = m_0.walls.L

    ax_vis.set_xlim(-L / 2.0, L / 2.0)
    ax_vis.set_ylim(-L / 2.0, L / 2.0)
    ax_vis.set_aspect('equal')

    plt.subplots_adjust(left=0.25, bottom=0.25)
    plot_p = ax_vis.quiver(m_0.r[:, 0], m_0.r[:, 1], m_0.v[:, 0], m_0.v[:, 1])

    wall_mask = np.ma.array(m_0.walls.a.T, mask=np.logical_not(m_0.walls.a.T))
    ax_vis.imshow(wall_mask, cmap='Greens_r', interpolation='nearest',
                  origin='lower', extent=2 * [-L / 2.0, L / 2.0])

    plot_c = ax_vis.imshow([[0]], cmap='Reds', interpolation='nearest',
                           origin='lower', extent=2 * [-L / 2.0, L / 2.0])

    ax_slide = plt.axes([0.25, 0.1, 0.65, 0.03])
    t_slider = Slider(ax_slide, 'Time', 0, len(fnames), valinit=0)

    ax_cb = plt.axes([0.875, 0.2, 0.05, 0.7])
    fig.colorbar(plot_c, cax=ax_cb)

    t_time = fig.text(0.1, 0.5, '')
    t_k = fig.text(0.1, 0.4, '')

    def update(val):
        fname_i = int(round(val))
        if 0 <= fname_i < len(fnames):
            m = output_utils.filename_to_model(fnames[fname_i])
            plot_p.set_offsets(m.r)
            plot_p.set_UVC(m.v[:, 0], m.v[:, 1])
            c_mask = np.ma.array(np.log(m.c.a.T),
                                 mask=np.logical_or(m.c.a.T == 0.0,
                                                    m.walls.a.T))
            plot_c.set_data(c_mask)
            plot_c.autoscale()
            t_time.set_text('Time: {:g}'.format(m.t))
            if m.n < 500:
                t_k.set_text('k: {:g}'.format(utils.get_k(m)))
            fig.canvas.draw_idle()

    t_slider.on_changed(update)

    plt.show()


def plot_1d(dirname):
    fig = plt.figure()
    ax_vis = fig.add_subplot(311)
    ax_c = fig.add_subplot(312)
    ax_d = fig.add_subplot(313)

    fnames = output_utils.get_filenames(dirname)

    m_0 = output_utils.filename_to_model(fnames[0])

    L = m_0.L

    ax_vis.set_xlim(-L / 2.0, L / 2.0)
    ax_c.set_xlim(-L / 2.0, L / 2.0)
    ax_d.set_xlim(-L / 2.0, L / 2.0)

    plt.subplots_adjust(left=0.25, bottom=0.25)
    plot_p = ax_vis.scatter(m_0.r[:, 0], np.zeros([m_0.n]))

    x = np.linspace(-m_0.L / 2.0, m_0.L / 2.0, m_0.c.a.shape[0])
    plot_c = ax_c.bar(x, m_0.c.a[:], width=x[1] - x[0])

    plot_d = ax_d.bar(x, m_0.c.get_density_field(), width=x[1] - x[0])

    ax_slide = plt.axes([0.25, 0.1, 0.65, 0.03])
    t_slider = Slider(ax_slide, 'Index', 0, len(fnames), valinit=0)

    def update(val):
        fname_i = int(round(val))
        if 0 <= fname_i < len(fnames):
            m = output_utils.filename_to_model(fnames[fname_i])
            plot_p.set_offsets(np.array([m.r[:, 0], np.zeros([m.n])]).T)
            cs = m.c.a
            for rect, c in zip(plot_c, cs):
                rect.set_height(c)
            ax_c.set_ylim(0.0, 1.05 * cs.max())
            ds = m.c.get_density_field() / m.rho_0
            for rect, d in zip(plot_d, ds):
                rect.set_height(d)
            ax_d.set_ylim(0.0, 1.05 * ds.max())
            fig.canvas.draw_idle()

    t_slider.on_changed(update)

    plt.show()


def plot_vis(dirname):
    dim = output_utils.get_recent_model(dirname).dim
    if dim == 1:
        plot_1d(dirname)
    elif dim == 2:
        plot_2d(dirname)


def plot_t_ks(dirname):
    fig = plt.figure()
    ax = fig.add_subplot(111)

    ts, ks = utils.t_ks(dirname)
    ax.plot(ts, ks)

    plt.show()


def plot_t_fracs(dirname):
    fig = plt.figure()
    ax = fig.add_subplot(111)

    ts, fracs = utils.t_fracs(dirname)
    for frac_set in fracs.T:
        ax.plot(ts, frac_set)

    plt.show()


def plot_t_pmeans(dirname):
    fig = plt.figure()
    ax = fig.add_subplot(111)

    ts, p_means, p_mins, p_maxs = utils.get_pmeans(dirname)
    ax.plot(ts, p_means)
    ax.plot(ts, p_mins)
    ax.plot(ts, p_maxs)

    plt.show()


def plot_chi_ks(dirnames):
    fig = plt.figure()
    ax = fig.add_subplot(111)

    chis, ks = utils.chi_ks(dirnames)
    i_sort = np.argsort(chis)
    chis, ks = chis[i_sort], ks[i_sort]
    ax.plot(chis, ks)
    ax.set_ylim(0.0, 1.1)

    plt.show()


def plot_chi_fs(dirnames):
    fig = plt.figure()
    ax = fig.add_subplot(111)

    chis, fracs = utils.chi_fs(dirnames)
    i_sort = np.argsort(chis)
    chis, fracs = chis[i_sort], fracs[i_sort]
    for frac_set in fracs.T:
        ax.plot(chis, frac_set)
    ax.set_ylim(0.0, 1.1)

    plt.show()


def vis_chi_k(dirnames, ax, label, c):
    m = utils.get_recent_model(dirnames[0])
    chis, ks = utils.chi_ks(dirnames, t_steady=None)
    # chis, fs = utils.chi_fs(dirnames, t_steady=None)
    i_sort = np.argsort(chis)
    chis, ks = chis[i_sort], ks[i_sort]
    # chis, fs = chis[i_sort], fs[i_sort]
    D_rhos = cutils.get_D_rho(m.v_0, m.p_0, m.dim)
    mus = cutils.get_mu(chis, m.v_0, m.p_0, m.onesided_flag, m.L)
    mus_red = cutils.get_reduced_mu(mus, m.c_source, m.rho_0, m.c_sink, m.c_D)
    D_rhos_red = cutils.get_reduced_D_rho(D_rhos, m.c_D)
    ax.plot(mus_red / D_rhos_red, ks, label=label, c=c)
    # ax.plot(chis, ks, label=label, c=c)
    # ax.plot(mus_red / D_rhos_red, fs, label=label, c=c)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
test_bannock
----------------------------------

Tests for `bannock` module.
"""

import unittest

from bannock import model


class TestBannock(unittest.TestCase):

    def test_bannock(self):
        m = model.Model1D(seed=1, dt=0.01, origin_flag=False,
                          rho_0=1.0, v_0=20.0, p_0=1.0,
                          chi=0.5, onesided_flag=True,
                          vicsek_R=1.0, c_D=1000.0, c_sink=0.01, c_source=1.0,
                          L=500.0, dx=50.0)
        m.iterate()
        self.assertTrue(True)



from __future__ import print_function, division
import numpy as np
import matplotlib.pyplot as plt


def find_c(rho,
           dx, dt,
           alpha, delta, D):
    c = np.zeros_like(rho)
    laplace_c = np.zeros_like(c)
    for _ in range(100000):
        laplace_c[1:-1] = (c[2:] + c[:-2] - 2.0 * c[1:-1]) / dx ** 2
        laplace_c[0] = 0.0
        laplace_c[-1] = 0.0
        c += (D * laplace_c + alpha * rho - delta * c) * dt
    return c

x = np.linspace(-2000.0, 2000.0, 800)
dx = x[1] - x[0]
delta = 0.01
alpha = 1.0
D = 100.0
d = 10.0 * np.exp(-0.01 * np.abs(x))

c = np.zeros_like(d)
for i in range(len(x)):
    for j in range(len(x)):
        c[i] += d[j] * np.exp(-np.sqrt(delta / D) * np.abs(x[i] - x[j]))
c *= 0.5 * dx

cs = find_c(d, dx=dx, dt=0.1, alpha=alpha, delta=delta, D=D)

plt.plot(x, d, label='density')
plt.plot(x, cs, label='c direct')
plt.plot(x, c, label='c analytic')

plt.legend()
plt.show()



from __future__ import print_function, division
import numpy as np
from ciabatta import vector, cell_list
from ciabatta.meta import make_repr_str
from bannock import numerics, walls
from bannock.secretion import Secretion, WalledSecretion


class AutoBaseModel(object):

    def __init__(self, dim, seed, dt,
                 origin_flag,
                 rho_0, v_0, p_0,
                 chi, onesided_flag,
                 vicsek_R,
                 c_D, c_sink, c_source):
        self.seed = seed
        self.dt = dt
        self.origin_flag = origin_flag
        self.v_0 = v_0
        self.p_0 = p_0
        self.chi = chi
        self.onesided_flag = onesided_flag
        self.vicsek_R = vicsek_R

        np.random.seed(seed)
        self.t = 0.0
        self.i = 0

    def _tumble(self):
        self.p[:] = self.p_0

        if self.chi:
            grad_c_i = self.c.grad_i(self.r)
            v_dot_grad_c = np.sum(self.v * grad_c_i, axis=-1)
            fitness = self.chi * v_dot_grad_c / self.v_0

            self.p *= 1.0 - fitness
            if self.onesided_flag:
                self.p = np.minimum(self.p_0, self.p)
            # self.p = np.maximum(self.p, 0.1)

        tumbles = np.random.uniform(size=self.n) < self.p * self.dt
        self.v[tumbles] = self.v_0 * vector.sphere_pick(self.dim,
                                                        tumbles.sum())

    def iterate(self):
        self.t += self.dt
        self.i += 1

    @property
    def L_half(self):
        return self.L / 2.0

    @property
    def n(self):
        return self.r.shape[0]

    @property
    def dim(self):
        return self.r.shape[1]

    @property
    def c_sink(self):
        return self.c.sink_rate

    @property
    def c_source(self):
        return self.c.source_rate

    @property
    def c_D(self):
        return self.c.D

    @property
    def dx(self):
        return self.c.D


class Model1D(AutoBaseModel):
    """Self-propelled particles moving in one dimension in a chemical field.

    Parameters
    ----------
    seed: int
        A random number seed. `None` causes a random choice.
    dt: float
        The size of a single time-step.
    origin_flag: bool
        Whether or not to start the particles at the centre of the system.
        If `True`, all particles are initialised in a small region near the
        origin.
        If `False`, particles are initialised uniformly.
    rho_0: float
        The average area density of particles.
    v_0: float
        The speed of the particles.
    p_0: float
        The base rate at which the particles randomise their
        direction.
    chi: float
        The sensitivity of the particles' chemotactic response to gradients
        in the chemoattractant concentration field.
    onesided_flag: bool
        Whether or not the particles' chemotactic response can increase
        their tumble rate.
    vicsek_R: float
        A radius within which the particles reorient with their neighbours.
    c_D: float
        see :class:`Secretion`
    c_sink: float
        see :class:`Secretion`
    c_source: float
        see :class:`Secretion`
    L: float
        Length of the system.
    dx: float
        Length of a cell in the chemical concentration field lattice.
    """

    def __init__(self, seed, dt,
                 origin_flag,
                 rho_0, v_0, p_0,
                 chi, onesided_flag,
                 vicsek_R,
                 c_D, c_sink, c_source,
                 L, dx,
                 *args, **kwargs):
        dim = 1
        super(Model1D, self).__init__(dim, seed, dt,
                                      origin_flag,
                                      rho_0, v_0, p_0,
                                      chi, onesided_flag,
                                      vicsek_R,
                                      c_D, c_sink, c_source)
        self.L = L

        n = int(round(self.L * rho_0))

        self.v = self.v_0 * vector.sphere_pick(dim, n)
        if self.origin_flag:
            # Randomise initialisation by a small distance, to avoid
            # unphysical regular spacing otherwise. In 1D the particles are
            # effectively on a lattice of length `v_0 * dt`.
            vdt = self.dt * self.v_0
            self.r = np.random.uniform(-vdt, vdt, [n, dim])
        else:
            self.r = np.random.uniform(-self.L_half, self.L_half, [n, dim])
        self.p = np.ones([self.n]) * self.p_0

        self.c = Secretion(self.L, dim, dx, c_D, self.dt, c_sink,
                           c_source, a_0=0.0)

    @property
    def rho_0(self):
        return self.n / self.L

    def _update_positions(self):
        self.r += self.v * self.dt
        self.r[self.r > self.L_half] -= self.L
        self.r[self.r < -self.L_half] += self.L

    def _vicsek(self):
        u = np.array(np.round(self.v[:, 0] / self.v_0), dtype=np.int)
        u_new = numerics.vicsek_1d(self.r[:, 0], u,
                                            self.vicsek_R, self.L)
        stats = u_new == 0
        u_new[stats] = 2 * np.random.randint(2, size=stats.sum()) - 1
        self.v[:, 0] = self.v_0 * u_new

    def iterate(self):
        """Evolve the model's state by a single time-step.

        - Do Vicsek alignment

        - Make particles tumble at their chemotactic probabilities.

        - Make the particles swim in the periodic space

        - Iterate the chemical concentration field
        """
        if self.vicsek_R:
            self._vicsek()
        if self.p_0:
            self._tumble()
        self._update_positions()

        if self.c_source:
            self.c.iterate(self.r)

        super(Model1D, self).iterate()

    def get_output_dirname(self):
        s = 'Bannock_{}D,seed={},dt={:g},'.format(self.dim, self.seed, self.dt)
        s += 'origin={:d},'.format(self.origin_flag)
        s += 'L={:g},'.format(self.L)
        s += 'rho={:g},v={:g},p={:g},'.format(self.rho_0, self.v_0, self.p_0)
        s += 'chi={:g},side={},'.format(self.chi, 2 - self.onesided_flag)
        s += 'vicsek_R={:g},'.format(self.vicsek_R)
        s += 'c_source={:g},c_sink={:g},c_D,dx={:g}'.format(self.c_source,
                                                            self.c_sink,
                                                            self.c_D,
                                                            self.dx)
        return s

    def __repr__(self):
        fs = [('dim', self.dim), ('seed', self.seed), ('dt', self.dt),
              ('origin_flag', self.origin_flag),
              ('L', self.L),
              ('n', self.n), ('v_0', self.v_0), ('p_0', self.p_0),
              ('chi', self.chi), ('onesided_flag', self.onesided_flag),
              ('vicsek_R', self.vicsek_R),
              ('c', self.c),
              ('t', self.t), ('i', self.i)
              ]
        return make_repr_str(self, fs)


class Model2D(AutoBaseModel):
    """Self-propelled particles moving in two dimensions in a chemical field.

    Parameters
    ----------
    D_rot: float
        The rotational diffusion constant of the particles.
    force_mu: float
        The degree to which the particles reorient towards
        :math:`\\nabla c`, where :math:`c` is the chemoattractant
        concentration field.
    walls: Walls
        Obstacles in the environment
    Others:
        see :class:`Model1D`.
    """

    def __init__(self, seed, dt,
                 origin_flag,
                 rho_0, v_0, p_0,
                 D_rot,
                 chi, onesided_flag,
                 vicsek_R,
                 force_mu,
                 c_D, c_sink, c_source,
                 walls,
                 *args, **kwargs):
        dim = 2
        super(Model2D, self).__init__(dim, seed, dt,
                                      origin_flag,
                                      rho_0, v_0, p_0,
                                      chi, onesided_flag,
                                      vicsek_R,
                                      c_D, c_sink, c_source)
        self.D_rot = D_rot
        self.force_mu = force_mu
        self.walls = walls

        self.c = WalledSecretion(self.walls.L, self.walls.dim, self.walls.dx,
                                 self.walls.a, c_D, self.dt,
                                 c_sink, c_source, a_0=0.0)

        n = int(round(self.walls.free_area * rho_0))

        self.v = self.v_0 * vector.sphere_pick(dim, n)
        self.r = np.zeros_like(self.v)
        if self.origin_flag:
            if self.walls.is_obstructed(self.r[0, np.newaxis]):
                raise Exception('Cannot initialise particles at the origin as '
                                'there is an obstacle there')
        else:
            for i in range(self.n):
                while True:
                    self.r[i] = np.random.uniform(-self.L_half, self.L_half,
                                                  dim)
                    if not self.walls.is_obstructed(self.r[i, np.newaxis]):
                        break
        self.p = np.ones([self.n]) * self.p_0

    @property
    def rho_0(self):
        return self.n / self.walls.free_area

    @property
    def L(self):
        return self.walls.L

    def _update_positions(self):
        r_old = self.r.copy()

        self.r += self.v * self.dt
        self.r[self.r > self.L_half] -= self.L
        self.r[self.r < -self.L_half] += self.L

        if self.walls.a.any():
            obs = self.walls.is_obstructed(self.r)
            # Find particles and dimensions which have changed cell.
            changeds = np.not_equal(self.walls.r_to_i(self.r),
                                    self.walls.r_to_i(r_old))
            # Find where particles have collided with a wall,
            # and the dimensions on which it happened.
            colls = np.logical_and(obs[:, np.newaxis], changeds)

            # Reset particle position components along which a collision
            # occurred
            self.r[colls] = r_old[colls]
            # Set velocity along that axis to zero.
            self.v[colls] = 0.0

            # Rescale new directions, randomising stationary particles.
            self.v[obs] = vector.vector_unit_nullrand(self.v[obs]) * self.v_0

    def _force(self):
        grad_c_i = self.c.grad_i(self.r)
        v_dot_grad_c = np.sum(self.v * grad_c_i, axis=-1)
        if self.onesided_flag:
            responds = v_dot_grad_c > 0.0
        else:
            responds = np.ones([self.n], dtype=np.bool)
        self.v[responds] += self.force_mu * grad_c_i[responds] * self.dt
        self.v = self.v_0 * vector.vector_unit_nullnull(self.v)

    def _rot_diff(self):
        self.v = numerics.rot_diff_2d(self.v, self.D_rot, self.dt)

    def _vicsek(self):
        inters, intersi = intro.get_inters(self.r, self.L, self.vicsek_R)
        self.v = numerics.vicsek_inters(self.v, inters, intersi)

    def iterate(self):
        """Evolve the model's state by a single time-step.

        - Do Vicsek alignment

        - Make particles tumble at their chemotactic probabilities.

        - Reorient particles according to chemotactic gradient

        - Diffuse the particles' directions

        - Make the particles swim in the periodic space

        - Reorient the particles that collide with walls, and move them back
          to their original positions

        - Iterate the chemical concentration field
        """
        if self.vicsek_R:
            self._vicsek()
        if self.p_0:
            self._tumble()
        if self.force_mu:
            self._force()
        if self.D_rot:
            self._rot_diff()
        self._update_positions()

        if self.c_source:
            self.c.iterate(self.r)
        super(Model2D, self).iterate()

    def get_output_dirname_walls_part(self):
        w = self.walls
        if w.__class__ == walls.Walls:
            return 'Walls'
        elif w.__class__ == walls.Traps:
            return 'Traps_n={},d={:g},w={:g},s={:g}'.format(w.n, w.d, w.w, w.s)
        elif w.__class__ == walls.Maze:
            return 'Maze_d={:g},seed={}'.format(w.d, w.seed)

    def get_output_dirname(self):
        s = 'Bannock_{}D,seed={},dt={:g},'.format(self.dim, self.seed, self.dt)
        s += 'origin={:d},'.format(self.origin_flag)
        s += 'walls={},'.format(self.get_output_dirname_walls_part())
        s += 'rho={:g},v={:g},p={:g},'.format(self.rho_0, self.v_0, self.p_0)
        s += 'Dr={:g},'.format(self.D_rot)
        s += 'chi={:g},side={},'.format(self.chi, 2 - self.onesided_flag)
        s += 'vicsek_R={:g},'.format(self.vicsek_R)
        s += 'force_mu={:g},'.format(self.force_mu)
        s += 'c_source={:g},c_sink={:g},c_D,dx={:g}'.format(self.c_source,
                                                            self.c_sink,
                                                            self.c_D,
                                                            self.dx)
        return s

    def __repr__(self):
        fs = [('dim', self.dim), ('seed', self.seed), ('dt', self.dt),
              ('origin_flag', self.origin_flag),
              ('walls', self.walls),
              ('n', self.n), ('v_0', self.v_0), ('p_0', self.p_0),
              ('D_rot', self.D_rot),
              ('chi', self.chi), ('onesided_flag', self.onesided_flag),
              ('vicsek_R', self.vicsek_R),
              ('force_mu', self.force_mu),
              ('c', self.c),
              ('t', self.t), ('i', self.i)
              ]
        return make_repr_str(self, fs)


class Model2DNoAlignment(Model2D):
    """Self-propelled particles moving in two dimensions in a chemical field.

    Particles do not align with obstacles, but instead reflect off.

    Parameters
    ----------
    see :class:`Model2D`.
    """

    def _update_positions(self):
        r_old = self.r.copy()

        self.r += self.v * self.dt
        self.r[self.r > self.L_half] -= self.L
        self.r[self.r < -self.L_half] += self.L

        if self.walls.a.any():
            obs = self.walls.is_obstructed(self.r)
            # Find particles and dimensions which have changed cell.
            changeds = np.not_equal(self.walls.r_to_i(self.r),
                                    self.walls.r_to_i(r_old))
            # Find where particles have collided with a wall,
            # and the dimensions on which it happened.
            colls = np.logical_and(obs[:, np.newaxis], changeds)

            # Reset particle position components along which a collision
            # occurred
            self.r[colls] = r_old[colls]
            # Reflect velocity along that axis.
            self.v[colls] = -self.v[colls]

            # Rescale new directions, randomising stationary particles.
            self.v[obs] = vector.vector_unit_nullrand(self.v[obs]) * self.v_0



from __future__ import print_function, division
from ciabatta import runner_utils
from agaro.run_utils import run_model
from ciabatta.parallel import run_func
from bannock.utils import utils


class _TaskResumeRunner(object):
    """Replacement for a closure, which I would use if
    the multiprocessing module supported them.

    Imagine `__init__` is the captured outside state,
    and `__call__` is the closure body.
    """

    def __init__(self, t_output_every, t_upto):
        self.t_output_every = t_output_every
        self.t_upto = t_upto

    def __call__(self, dirname):
        m = runner_utils.get_recent_model(dirname)
        r = run_model(self.t_output_every, m=m,
                      force_resume=True, t_upto=self.t_upto)


def run_dirname_scan_resume(dirnames, t_output_every, t_upto,
                            parallel=False):
    """Run many models to a later time.

    For each `dirname` in `dirnames`, a model will be resumed, and run up to a
    time.

    Parameters
    ----------
    dirnames: list[str]
        Output directories to resume
    t_output_every: float
        see :class:`Runner`.
    t_upto: float
        Run each model until the time is equal to this
    parallel: bool
        Whether or not to run the models in parallel, using the Multiprocessing
        library. If `True`, the number of concurrent tasks will be equal to
        one less than the number of available cores detected.
     """
    task_runner = _TaskResumeRunner(t_output_every, t_upto)
    run_func(task_runner, dirnames, parallel)



from __future__ import print_function, division
import numpy as np
from ciabatta import fields
from ciabatta.meta import make_repr_str


class Secretion(fields.Diffusing):
    """A concentration field of a secreted, diffusing, decaying chemical.

    Parameters
    ----------
    L: float
        The length of the field.
    dim: int
        The number of dimensions.
    dx: float
        The length of a cell.
    D: float
        The diffusion constant of the secreted chemical.
    dt: float
        The time-step represented by one iteration.
    sink_rate: float
        The fraction of the chemical which decays per unit time.
        Units are inverse time.
    source_rate: float
        The increase in secretion concentration per unit time,
        per unit secreter density.
    a_0: array_like
        Initial field state.
    """

    def __init__(self, L, dim, dx, D, dt, sink_rate, source_rate,
                 a_0=0.0):
        fields.Diffusing.__init__(self, L, dim, dx, D, dt,
                                  a_0=a_0)
        self.source_rate = source_rate
        self.sink_rate = sink_rate

    def iterate(self, r):
        """
        Evolve the field's state according to its differential equation, by a
        single time-step.

        Parameters
        ----------
        density: numpy.ndarray[dtype=float]
            The density of secreter.
        """
        density = self.density_field(r)
        fields.Diffusing.iterate(self)
        self.a += self.dt * (self.source_rate * density -
                             self.sink_rate * self.a)
        self.a = np.maximum(self.a, 0.0)

    def __repr__(self):
        fs = [('L', self.L), ('dim', self.dim), ('dx', self.dx),
              ('D', self.D), ('dt', self.dt), ('sink_rate', self.sink_rate),
              ('source_rate', self.source_rate),
              ]
        return make_repr_str(self, fs)


class WalledSecretion(fields.WalledDiffusing):
    """A concentration field of a secreted chemical in a walled environment.

    Parameters
    ----------
    walls: numpy.ndarray[dtype=bool]
        An array of the same shape as the field,
        where `True` indicates the presence of an obstacle.
    Others: see :class:`Secretion`.
    """

    def __init__(self, L, dim, dx, walls, D, dt, sink_rate, source_rate,
                 a_0=0.0):
        fields.WalledDiffusing.__init__(self, L, dim, dx, walls, D, dt,
                                        a_0=a_0)
        self.source_rate = source_rate
        self.sink_rate = sink_rate

    def iterate(self, r):
        """See :meth:`Secretion.iterate`."""
        density = self.density_field(r)
        fields.WalledDiffusing.iterate(self)
        self.a += self.dt * (self.source_rate * density -
                             self.sink_rate * self.a)
        self.a = np.maximum(self.a, 0.0)

    def __repr__(self):
        fs = [('L', self.L), ('dim', self.dim), ('dx', self.dx),
              ('D', self.D), ('dt', self.dt), ('sink_rate', self.sink_rate),
              ('source_rate', self.source_rate)]
        return make_repr_str(self, fs)



from __future__ import print_function, division
import numpy as np
from ciabatta import cluster
from agaro.output_utils import (get_recent_filename, get_filenames,
                                filename_to_model)
from agaro.measure_utils import get_average_measure, group_by_key


r_cluster_1d = 5.0
r_cluster_2d = 20.0


def _get_r_cluster(dim):
    """Find the cluster length scale appropriate for a model's dimension.

    Parameters
    ----------
    dim: int
        Dimension.

    Returns
    -------
    r_cluster: float
        Cluster length scale.
    """
    if dim == 1:
        return r_cluster_1d
    elif dim == 2:
        return r_cluster_2d


def get_k(m):
    """Calculate the particle clumpiness for a model.

    Parameters
    ----------
    m: Model
        Model instance.

    Returns
    -------
    k: float
        Clumpiness measure.
    """
    labels = cluster.cluster_periodic(m.r, _get_r_cluster(m.dim), m.L)
    clust_sizes = cluster.cluster_sizes(labels)
    return cluster.get_clumpiness(clust_sizes)


def get_fracs(m):
    """Calculate an order parameter, f, the 'confinedness',
    representing the fraction of particles in each trap for a model.

    f = 1 means all particles are in a trap.
    f = 0 means the fraction is that expected for a uniform distribution.
    f < 0 means there are fewer particles than expected.

    Parameters
    ----------
    m: Model
        Model instance.

    Returns
    -------
    fs: list[float]
        Confinedness for each trap.
    """
    fracs_0 = m.walls.get_trap_areas() / m.walls.get_free_area()
    return (m.walls.get_fracs(m.r) - fracs_0) / (1.0 - fracs_0)


def get_pstats(m):
    """Calculate the tumble rate statistics for a model.

    Parameters
    ----------
    m: Model
        Model instance.

    Returns
    -------
    p_mean: float
        Mean tumble rate, using a floor of zero, so that 0, -1 and -100
        all count as 0 when calculating the mean. This is done even if
        the model does not do this.
    p_mean: float
        Minimum tumble rate. A floor of zero is *not* used.
    p_max:float
        Maximum tumble rate.
    """
    return np.maximum(m.p, 0.0).mean(), m.p.min(), m.p.max()


def t_ks(dirname):
    """Calculate the particle clumpiness over time
    for a model output directory.

    Parameters
    ----------
    dirname: str
        A model output directory path

    Returns
    -------
    ts: numpy.ndarray[dtype=float]
        Times.
    ks: numpy.ndarray[dtype=float]
        Particle clumpinesses.
    """
    ts, ks = [], []
    for fname in get_filenames(dirname):
        m = filename_to_model(fname)
        ts.append(m.t)
        ks.append(get_k(m))
    return np.array(ts), np.array(ks)


def t_fracs(dirname):
    """Calculate the trap confinedness over time
    for a model output directory.

    Parameters
    ----------
    dirname: str
        A model output directory path

    Returns
    -------
    ts: numpy.ndarray[dtype=float]
        Times.
    fs: numpy.ndarray[dtype=float, shape=(ts.shape[0], num_traps)]
        Trap confinednesses, where `num_traps` is the number of traps.
    """
    ts, fracs = [], []
    for fname in get_filenames(dirname):
        m = filename_to_model(fname)
        ts.append(m.t)
        fracs.append(get_fracs(m))
    return np.array(ts), np.array(fracs)


def t_pmeans(dirname):
    """Calculate tumble rates statistics over time for a model output directory.

    Parameters
    ----------
    dirname: str
        A model output directory path

    Returns
    -------
    ts: numpy.ndarray[dtype=float]
        Times.
    p_means: numpy.ndarray[dtype=float]
        Mean tumble rates, using a floor of zero, so that 0, -1 and -100 all
        count as 0 when calculating the mean. This is done even if the model
        does not do this.
    p_mins: numpy.ndarray[dtype=float]
        Minimum tumble rates. A floor of zero is *not* used.
    p_maxs: numpy.ndarray[dtype=float]
        Maximum tumble rates.
    """
    ts, p_means, p_mins, p_maxs = [], [], [], []
    for fname in get_filenames(dirname):
        m = filename_to_model(fname)
        ts.append(m.t)
        p_mean, p_min, p_max = get_pstats(m)
        p_means.append(p_mean)
        p_mins.append(p_min)
        p_maxs.append(p_max)
    return np.array(ts), np.array(p_means), np.array(p_mins), np.array(p_maxs)


def _chi_measures(dirnames, measure_func, t_steady=None):
    """Calculate a measure of a set of
    model output directories, and their associated chis.

    Parameters
    ----------
    dirnames: list[str]
        Model output directory paths.
    t_steady: None or float
        Time to consider the model to be at steady-state.
        The measure will be averaged over all later times.
        `None` means just consider the latest time.

    Returns
    -------
    chis: numpy.ndarray[dtype=float]
        Chemotactic sensitivities
    measures: numpy.ndarray[dtype=float]
        Measures.
    """
    chis, measures = [], []
    for dirname in dirnames:
        m_recent = filename_to_model(get_recent_filename(dirname))
        chis.append(m_recent.chi)
        measures.append(get_average_measure(dirname, measure_func, t_steady))
    return np.array(chis), np.array(measures)


def chi_ks(dirnames, t_steady=None):
    """Calculate the particle clumpiness of a set of
    model output directories, and their associated chis.

    Parameters
    ----------
    dirnames: list[str]
        Model output directory paths.
    t_steady: None or float
        Time to consider the model to be at steady-state.
        The measure will be averaged over all later times.
        `None` means just consider the latest time.

    Returns
    -------
    chis: numpy.ndarray[dtype=float]
        Chemotactic sensitivities
    ks: numpy.ndarray[dtype=float]
        Particle clumpinesses.
    """
    return _chi_measures(dirnames, get_k, t_steady)


def chi_fs(dirnames, t_steady=None):
    """Calculate particle confinedness of a set of
    model output directories, and their associated chis.

    Parameters
    ----------
    dirnames: list[str]
        Model output directory paths.
    t_steady: None or float
        Time to consider the model to be at steady-state.
        The measure will be averaged over all later times.
        `None` means just consider the latest time.

    Returns
    -------
    chis: numpy.ndarray[dtype=float]
        Chemotactic sensitivities
    fs: numpy.ndarray[dtype=float]
        Particle confinednesses.
    """
    return _chi_measures(dirnames, get_fracs, t_steady)


def chi_ks_run_average(dirnames, t_steady=None):
    """Calculate the particle clumpiness of a set of
    model output directories, and their associated chis.
    Take the average over all directories with equal chi.

    Parameters
    ----------
    dirnames: list[str]
        Model output directory paths.
    t_steady: None or float
        Time to consider the model to be at steady-state.
        The measure will be averaged over all later times.
        `None` means just consider the latest time.

    Returns
    -------
    chis: numpy.ndarray[dtype=float]
        Chemotactic sensitivities
    ks: numpy.ndarray[dtype=float]
        Particle clumpinesses.
    """
    chi_groups = group_by_key(dirnames, 'chi')
    chis, ks = [], []
    for chi, dirnames in chi_groups.items():
        chis.append(chi)
        ks.append(np.mean([get_average_measure(dirname, get_k, t_steady)
                           for dirname in dirnames]))
    return np.array(chis), np.array(ks)



from __future__ import print_function, division
import numpy as np
from ciabatta import lattice, fields, maze
from ciabatta.meta import make_repr_str


class Walls(fields.Field):
    """A field representing an environment filled with square obstacles."""

    def __init__(self, L, dim, dx):
        fields.Field.__init__(self, L, dim, dx)
        self.a = np.zeros(self.dim * (self.M,), dtype=np.uint8)

    @property
    def free_area_i(self):
        """Calculate the number of elements that are not occupied by obstacles.

        Returns
        -------
        area_i: int
        """
        return np.logical_not(self.a).sum()

    @property
    def free_area(self):
        """Calculate the area that is not occupied by obstacles.

        Returns
        -------
        area: float
        """
        return self.free_area_i * self.dA

    def is_obstructed(self, r):
        """Determine if a set of position vectors lie on top of obstacles.

        Parameters
        ----------
        r: array_like[shape=(n, 2)]
            Particle position vectors.

        Returns
        -------
        o: numpy.ndarray[dtype=bool, shape=(n,)]
            `True` if a vector is obstructed.
        """
        return self.a[tuple(self.r_to_i(r).T)].astype(np.bool)

    def __repr__(self):
        fs = [('L', self.L), ('dim', self.dim), ('dx', self.dx)]
        return make_repr_str(self, fs)


class HalfClosed(Walls):
    """A set of walls closing the 2D environment along one edge."""

    def __init__(self, L, dx):
        Walls.__init__(self, L, dim=2, dx=dx)
        self.a[:, 0] = True

    def __repr__(self):
        fs = [('L', self.L), ('dx', self.dx)]
        return make_repr_str(self, fs)


class Closed(Walls):
    """A set of walls closing the 2D environment at all edges."""

    def __init__(self, L, dx):
        Walls.__init__(self, L, dim=2, dx=dx)
        self.a[:, 0] = True
        self.a[:, -1] = True
        self.a[0, :] = True
        self.a[-1, :] = True

    def __repr__(self):
        fs = [('L', self.L), ('dx', self.dx)]
        return make_repr_str(self, fs)


class Tittled(Walls):

    def __init__(self, L, dx, wx, wy, sx, sy):
        Walls.__init__(self, L, dim=2, dx=dx)
        self.wx_i = int(round(wx / self.dx))
        self.wy_i = int(round(wy / self.dx))
        self.sx_i = int(round(sx / self.dx))
        self.sy_i = int(round(sy / self.dx))

        for i_x in range(self.sx_i,
                         self.a.shape[0] - self.sx_i,
                         self.sx_i):
            for i_y in range(self.sy_i,
                             self.a.shape[1] - self.sy_i,
                             self.sy_i):
                self.a[i_x - self.wx_i:i_x + self.wx_i,
                       i_y - self.wy_i:i_y + self.wy_i] = True

    @property
    def wx(self):
        return self.wx_i * self.dx

    @property
    def wy(self):
        return self.wy_i * self.dx

    @property
    def sx(self):
        return self.sx_i * self.dx

    @property
    def sy(self):
        return self.sy_i * self.dx

    def __repr__(self):
        fs = [('L', self.L), ('dx', self.dx),
              ('wx', self.wx), ('wy', self.wy),
              ('sx', self.sx), ('sy', self.sy)]
        return make_repr_str(self, fs)


class Traps(Walls):
    """A set of walls forming a number of 2D traps.

    Parameters
    ----------
    n: int
        The number of traps. Can be 1, 4 or 5.
    d: float
        The width of the trap wall.
        Valid values are `i * dx`, where `i` is an integer >= 1.
    w: float
        The width of the entire trap.
        Valid values are `(2i + 1) dx`, where `i` is an integer >= 0.
    s: float
        The width of the trap entrance.
        Valid values are `(2i + 1) dx`, where `i` is an integer >= 0.
    """

    def __init__(self, L, dx, n, d, w, s):
        Walls.__init__(self, L, dim=2, dx=dx)

        # Calculate length in terms of lattice indices
        d_i = int(round(d / self.dx))
        w_i = int(round(w / self.dx))
        s_i = int(round(s / self.dx))

        # Calculate how many indices to go in each direction
        w_i_half = w_i // 2
        s_i_half = s_i // 2
        # l is the width of the trap, including its walls.
        l_i_half = w_i_half + d_i
        # Going to carve out `w_i_half` in each direction from a cell,
        # so will carve out this many cells.
        w_i = 2 * w_i_half + 1
        # Same goes for the slit.
        s_i = 2 * s_i_half + 1

        self.d_i = d_i
        self.w_i = w_i
        self.s_i = s_i

        if self.w < 0.0 or self.w > self.L:
            raise Exception('Invalid trap width')
        if self.s < 0.0 or self.s > self.w:
            raise Exception('Invalid slit length')

        if n == 1:
            traps_f = np.array([[0.50, 0.50]],
                               dtype=np.float)
        elif n == 4:
            traps_f = np.array([[0.25, 0.25], [0.25, 0.75], [0.75, 0.25],
                                [0.75, 0.75]],
                               dtype=np.float)
        elif n == 5:
            traps_f = np.array([[0.25, 0.25], [0.25, 0.75],
                                [0.50, 0.50],
                                [0.75, 0.25], [0.75, 0.75]],
                               dtype=np.float)
        else:
            raise Exception('Traps not implemented for this number of traps')

        self.traps_i = np.asarray(self.M * traps_f, dtype=np.int)
        for x, y in self.traps_i:
            # First fill in entire trap-related area.
            self.a[x - l_i_half:x + l_i_half + 1,
                   y - l_i_half:y + l_i_half + 1] = True
            # Then carve out trap interior.
            self.a[x - w_i_half:x + w_i_half + 1,
                   y - w_i_half:y + w_i_half + 1] = False
            # Then make the slit.
            self.a[x - s_i_half:x + s_i_half + 1,
                   y + w_i_half:y + l_i_half + 1] = False

    @property
    def d(self):
        return self.d_i * self.dx

    @property
    def w(self):
        return self.w_i * self.dx

    @property
    def s(self):
        return self.s_i * self.dx

    @property
    def n(self):
        return self.traps_i.shape[0]

    def get_trap_areas_i(self):
        """Calculate the number of elements occupied by each trap.

        Returns
        -------
        trap_areas_i: list[int]
        """
        trap_areas_i = []
        w_i_half = self.w_i // 2
        for x, y in self.traps_i:
            trap = self.a[x - w_i_half:x + w_i_half + 1,
                          y - w_i_half:y + w_i_half + 1]
            trap_areas_i.append(np.sum(np.logical_not(trap)))
        return np.array(trap_areas_i)

    def get_trap_areas(self):
        """Calculate the area occupied by each trap.

        Returns
        -------
        trap_areas: list[float]
        """
        return self.get_trap_areas_i() * self.dA

    def get_fracs(self, r):
        """Calculate the fraction of particles inside each trap.

        Parameters
        ----------
        r: array_like[shape=(n, 2)]
            Particle position vectors.

        Returns
        -------
        fracs: list[int]
            Fraction of the total population that is inside each trap.
        """
        inds = self.r_to_i(r)
        n_traps = [0 for i in range(len(self.traps_i))]
        w_i_half = self.w_i // 2
        for i_trap in range(len(self.traps_i)):
            mid_x, mid_y = self.traps_i[i_trap]

            low_x, high_x = mid_x - w_i_half, mid_x + w_i_half
            low_y, high_y = mid_y - w_i_half, mid_y + w_i_half
            for i_x, i_y in inds:
                if low_x <= i_x <= high_x and low_y <= i_y <= high_y:
                    n_traps[i_trap] += 1
        return np.array([float(n_trap) / float(r.shape[0])
                         for n_trap in n_traps])

    def __repr__(self):
        fs = [('L', self.L), ('dx', self.dx),
              ('n', self.n), ('d', self.d), ('w', self.w), ('s', self.s)]
        return make_repr_str(self, fs)


class Maze(Walls):
    """A set of walls forming a maze.

    Parameters
    ----------
    d: float
        The width of the maze walls.
    seed: int
        The random number seed used to generate the maze.
        Note that this does not affect, or is affected by, pre-existing
        random number seeding.
    """

    def __init__(self, L, dim, dx, d, seed=None):
        Walls.__init__(self, L, dim, dx)
        self.seed = seed
        rng = np.random.RandomState(self.seed)
        self.M_m = int(round(self.L / d))
        self.d_i = int(round(self.M / self.M_m))
        a_base = maze.make_maze_dfs(self.M_m, self.dim, rng)
        self.a[...] = lattice.extend_array(a_base, self.d_i)

    @property
    def d(self):
        return self.d_i * self.dx

    def __repr__(self):
        fs = [('L', self.L), ('dim', self.dim), ('dx', self.dx), ('d', self.d),
              ('seed', self.seed)]
        return make_repr_str(self, fs)



import numpy as np
from ciabatta import vector
cimport numpy as np
cimport cython
from libc.math cimport sin, cos


@cython.cdivision(True)
@cython.boundscheck(False)
def vicsek_inters(np.ndarray[np.float_t, ndim=2] v,
                  np.ndarray[int, ndim=2] inters,
                  np.ndarray[int, ndim=1] intersi):
    cdef:
        unsigned int i_1, i_i_2, i_dim
        np.ndarray[np.float_t, ndim=2] v_vic = v.copy()

    for i_1 in range(v.shape[0]):
        for i_i_2 in range(intersi[i_1]):
            for i_dim in range(v.shape[1]):
                v_vic[i_1, i_dim] += v[inters[i_1, i_i_2] - 1, i_dim]
    return vector.vector_unit_nullnull(v_vic) * vector.vector_mag(v)[:, np.newaxis]


@cython.cdivision(True)
@cython.boundscheck(False)
def vicsek_1d(np.ndarray[np.float_t, ndim=1] r,
              np.ndarray[np.int_t, ndim=1] u,
              double r_v, double L):
    cdef:
        unsigned int i_1, i_2
        np.ndarray[np.int_t, ndim=1] u_vic = np.zeros([u.shape[0]], dtype=np.int)
        double dr

    for i_1 in range(u.shape[0]):
        for i_2 in range(u.shape[0]):
            dr = abs(r[i_1] - r[i_2])
            dr = min(dr, L - dr)
            if dr < r_v:
                u_vic[i_1] += u[i_2]
        if u_vic[i_1] > 0:
            u_vic[i_1] = 1
        elif u_vic[i_1] < 0:
            u_vic[i_1] = -1
        else:
            u_vic[i_1] = 0
    return u_vic


@cython.cdivision(True)
@cython.boundscheck(False)
def rot_diff_2d(np.ndarray[np.float_t, ndim=2] v, double D, double dt):
    cdef:
        unsigned int i
        np.ndarray[np.float_t, ndim=1] th = np.random.normal(scale=np.sqrt(2.0 * D * dt), size=v.shape[0])
        np.ndarray[np.float_t, ndim=2] v_rot = np.empty((v.shape[0], v.shape[1]))
        double cos_th, sin_th

    for i in range(v.shape[0]):
        cos_th = cos(th[i])
        sin_th = sin(th[i])
        v_rot[i, 0] = cos_th * v[i, 0] - sin_th * v[i, 1]
        v_rot[i, 1] = sin_th * v[i, 0] + cos_th * v[i, 1]
    return v_rot
